/*
   - The contents of this file are subject to the Mozilla Public
   - License Version 1.1 (the "License"); you may not use this file
   - except in compliance with the License. You may obtain a copy of
   - the License at http://www.mozilla.org/MPL/
   -
   - Software distributed under the License is distributed on an "AS
   - IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
   - implied. See the License for the specific language governing
   - rights and limitations under the License.
   -
   - The Original Code is this file as it was released on
   - August 21, 2003.
   -
   - The Initial Developer of the Original Code is HJ van Rantwijk.
   - Portions created by HJ van Rantwijk are Copyright (C) 2003-2005
   - HJ van Rantwijk.  All Rights Reserved.
   -
   - Contributor(s):
   -	Leland M. George <Leland.M.George@alum.mit.edu> (bug 9547)
*/

const TAB_BUTTON        = 0;
const WINDOW_BUTTON     = 1;
const SESSION_BUTTON    = 2;
const INTERVAL_BUTTON   = 3;
const PROPERTIES_BUTTON = 4;
const CLOSE_BUTTON      = 5;
const QPREFS_BUTTON     = 6;

const CM_SELECT_ALL  = 0;
const CM_SELECT_NONE = 1;
const CM_INTERVAL    = 2;

const NO_ROWS_AVAILABLE      = -1;
const NO_ROWS_SELECTED       = 0;
const SINGLE_ROW_SELECTED    = 1;
const MULTIPLE_ROWS_SELECTED = 2;
const ROW_IS_CONTAINER       = 4;
const MIXED_ROWS_SELECTED    = 8;

const panelPath = "chrome://multiviews/content/tab-manager/preferences/"; // tabManagerPrefWindow.js
const gWindowManager = Components.classes['@mozilla.org/appshell/window-mediator;1'].getService(Components.interfaces.nsIWindowMediator);
const nsIWindowDataSource = Components.interfaces.nsIWindowDataSource;
const gWindowManagerDS = Components.classes['@mozilla.org/rdf/datasource;1?name=window-mediator'].getService(nsIWindowDataSource);
const gObserverService = Components.classes['@mozilla.org/observer-service;1'].getService(Components.interfaces.nsIObserverService);
const gPromptService = Components.classes["@mozilla.org/embedcomp/prompt-service;1"].getService(Components.interfaces.nsIPromptService);
  
/***
  * This observer is triggered by the 'DOMAttrModified' event handler on tab's
  * see also: <binding name="tab"... in multiviews.xml
  */
const gTabUpdateObserver = 
{
  observe:function(subject, topic, data) {
    /***
      * Prevent event bubbling for 'selected' events to refresh our tree
      */ 
    if (data == "selected") {
      if (subject.ordinal == gLastTabOrdinal)
        return;
      else 
        gLastTabOrdinal = subject.ordinal;
    }
    setTimeout(tabManager.refreshTree, 0);
  }
};
/***
  * This listener is where all the magic happens ;)
  */
const gWindowManagerListener = 
{
  onOpenWindow: function(aXULWindow) {
  },

  onCloseWindow: function(aXULWindow) {
    sanatizeTree(aXULWindow);
  },

  onWindowTitleChange: function(aXULWindow, aWindowTitle) {
    var chromeWindow = getChromeWindowFromXULWindow(aXULWindow);
    var windowID = getWindowResourceID(chromeWindow);

    if (windowID) {
      var windowNumber = windowID.replace(/window-/, '');

      if (windowNumber > gHigestWindowNumber)
        gHigestWindowNumber = windowNumber;

      chromeWindow = gWindowManagerDS.getWindowForResource(windowID);

      if (isBrowserWindow(chromeWindow)) {
        var targetResource = "urn:" + windowID + ":tabs:root";
        var rows = tree.treeBoxObject.view.rowCount;

        for (var row = 0; row < rows; row++) {
          if (tree.treeBoxObject.view.isContainer(row)) {
            var resource = tabManager.getResourceForRow(row);

            if (resource == targetResource) {
              targetResource = null;
              break;
            }
          }
        }
        if (targetResource) {
          var found = false;
          var browser = chromeWindow.getBrowser();

          for (i in gWindowResources) {
            if (gWindowResources[i] == targetResource) {
              found = true;
              break;
            }
          }
          if (!found) {
            gWindowResources.push(targetResource);
            gObserverService.notifyObservers(window.document.documentElement, "MZManagers", "start");
          }
          storeGroupData(dataSource, targetResource, browser, true, false, gSelectedTabStyle);
        }
      }  
    }
    /* We might want to replace the urn: with the window title some day, however, 
       it isn't implemented yet, because there's too much other stuff to fix first!

    var chromeWindow = getChromeWindowFromXULWindow(aXULWindow);

    if (isBrowserWindow(chromeWindow)) {
      var windowID = getWindowResourceID(chromeWindow);
      var targetResource = "urn:" + windowID + ":tabs:root";
      writeSessionName(dataSource, targetResource, aWindowTitle);
    } */
  }
}

var tree = null;
var pref = null; // declared in navigator.js but out of reach for us!
var dataSource = null;
var prefWindow = null;
var prefPanelNames = "";
var dialogType = "tabManagerPrefWindow";
var managerTypes = [ "tabManager" ];
// var resourcePrefix = "http://multizilla.mozdev.org/rdf#"; // this one doesn't seem to be used anymore!

/***
  * Toolbar button ID arrays that will be replaced with elements in updateToolbarButtons()
  * note: keep the order in-line with the corresponding XUL code!
  */
var toolbarButtons = [ "tabButton", "windowButton", "sessionButton", "reloadButton", "propertiesButton", "closeButton", "prefsButton" ];
/***
  * Context Menu ID array that will be replaced with elements in initContextMenu(aEvent)
  * note: we have to keep the order in-line with the corresponding XUL code!
  */
var contextMenuItems = [ "doCmd_SelectAll", "doCmd_UndoSelection", "SetInterval" ];
/***
  * Toolbar button arrays, to ease the pain of enabling/disabling the buttons
  */
var defaultEnabledButtons      = [  true,  true, false, false, false, false, true ];
var selectionEnabledButtons    = [  true,  true,  true,  true,  true,  true, true ];
var multipleRowsEnabledButtons = [ false,  true, false, false, false,  true, true ];

var busyTabs = -1;

var gResource = "";
var gTabbrowser = null;
var gWindowResources = new Array();
var gUseStackedWindows = false;
var gHigestWindowNumber = -1;
var gLastTabOrdinal = -1;
/***
  * Keep this in sync with gSelectedTabStyle in tabManagerMainPanel.js
  */
var gSelectedTabStyle = "bold"; // "none", "italic", "bold", "italic-and-bold", "red", "italic-and-red"
var gStringBundle = null;

// Global vars used by confirmation dialog
var gConfirmCheckbox = false;
var gEscapeKeyPressed = false;
var gConfirmForID = 0;

////////////////////////////////////////////////////////////////////////////////
function selectWindowStyling(aWindow)
{
  var documentElement = aWindow.document.documentElement;

  if (window.opener.mzPreferences.readMyPref("multizilla.windows.classic-style", "bool", false))
    documentElement.removeAttribute("extraStyling");
  else
    documentElement.setAttribute("extraStyling", "true");
}
////////////////////////////////////////////////////////////////////////////////
function sanatizeTree(aXULWindow)
{
  if (isBrowserWindow(aXULWindow)) {
    var rows = tree.treeBoxObject.view.rowCount;
    tree.view.selection.clearSelection();
    tree.currentIndex = -1;

    for (var row = 0; row < rows; row++) {
      if (tree.treeBoxObject.view.isContainer(row)) {
        var resource = tabManager.getResourceForRow(row);
        var windowID = resource.replace(/urn:/, '').replace(/:tabs:root/, '');
        var targetWindow = gWindowManagerDS.getWindowForResource(windowID);

        if (targetWindow == null || isBrowserWindow(targetWindow) == false) {
          gWindowResources.pop(resource);
          removeAllResourcesFor(dataSource, resource); // remove window from tree
          rows = tree.treeBoxObject.view.rowCount;
        }
      }
    }
  }
}
////////////////////////////////////////////////////////////////////////////////
function getChromeWindowFromXULWindow(aXULWindow)
{
  var chromeWindow = null;
  var docShell = aXULWindow.docShell;

  if (docShell) {
    var requestor = docShell.QueryInterface(Components.interfaces.nsIInterfaceRequestor);
    chromeWindow = requestor.getInterface(Components.interfaces.nsIDOMWindow);
  }
  return chromeWindow;
}
////////////////////////////////////////////////////////////////////////////////
function isBrowserWindow(aWindow)
{
  var chromeWindow = (aWindow instanceof ChromeWindow) ? aWindow : getChromeWindowFromXULWindow(aWindow);

  if (chromeWindow && (chromeWindow.document.documentElement.getAttribute('windowtype') == 'navigator:browser'))
    return true;
  return false;
}
////////////////////////////////////////////////////////////////////////////////
function getXULWindowResourceID(aXULWindow)
{
  var retValue = -1;
  var chromeWindow = getChromeWindowFromXULWindow(aXULWindow);

  if (chromeWindow && isBrowserWindow(chromeWindow))
    retValue = getWindowResourceID(chromeWindow);

  return retValue;
}
////////////////////////////////////////////////////////////////////////////////
function getNavigatorWindow(aOpenFlag)
{
  var windowManager = Components.classes['@mozilla.org/appshell/window-mediator;1'].getService();
  var windowManagerInterface = windowManager.QueryInterface(Components.interfaces.nsIWindowMediator);
  var navigatorWindow = windowManagerInterface.getMostRecentWindow("navigator:browser");

  if (!navigatorWindow) {
    if (!aOpenFlag)
      return null;
    else {
      window.open();
      navigatorWindow = windowManagerInterface.getMostRecentWindow("navigator:browser");
    }
  }
  return navigatorWindow;
}
////////////////////////////////////////////////////////////////////////////////
function getTabbrowser(aOpenFlag)
{
  var navigatorWindow = getNavigatorWindow(aOpenFlag);

  if (navigatorWindow)
    return navigatorWindow.getBrowser();
  return null;
}
////////////////////////////////////////////////////////////////////////////////
function areAllTabsBlank(aTabbrowser)
{
  var numberOfTabs = aTabbrowser.mTabContainer.childNodes.length;
  var numberOfBlankTabs = getNumberOfBlankTabs(aTabbrowser);

  return (numberOfTabs == numberOfBlankTabs);
}
////////////////////////////////////////////////////////////////////////////////
function getNumberOfBlankTabs(aTabBrowser)
{
  var blankTabs = 0;
  var browsers = aTabBrowser.mPanelContainer.childNodes.length;

  for (var i = 0; i < browsers; i++) {
    var browser = aTabBrowser.getBrowserAtIndex(i);
    var tab = aTabBrowser.getTabForBrowser(browser);

    if (tab.hasAttribute("busy")) {
      if (tab.getAttribute("firstPage") == "about:blank")
        blankTabs++;
    }
    else if (browser.currentURI.spec == "about:blank")
      blankTabs++;
  }
  return blankTabs;
}
////////////////////////////////////////////////////////////////////////////////
function getBrowser()
{
  return getTabbrowser(true);
}
////////////////////////////////////////////////////////////////////////////////
// This function is copied over from multiviewsInit.js because it is not 
// available from multiviewsRDF.js (read 'it is not included').
////////////////////////////////////////////////////////////////////////////////
function getNumberOfWindowsByType(aType)
{
  var defaultType = "navigator:browser";

  if (!aType) // default check for browser window
    aType = defaultType;

  var windowCounter = 0;
  var enumerator = gWindowManager.getEnumerator(aType);

  while (enumerator.hasMoreElements()) {
    var aWindow = enumerator.getNext();
    windowCounter++;
  }
  return windowCounter;
}
////////////////////////////////////////////////////////////////////////////////
function updateToolbarButtons()
{
  // Is the toolbar button array initialized?
  if (typeof toolbarButtons[0] == 'string') {
    // Nope, not yet, initialize array now
    for (i in toolbarButtons) {
      toolbarButtons[i] = document.getElementById(toolbarButtons[i]);
    }
  }
  var buttons = defaultEnabledButtons;
  var treeState = tabManager.getCurrentTreeState();

  if (treeState > NO_ROWS_SELECTED) {
    if (treeState & SINGLE_ROW_SELECTED) {
      buttons = selectionEnabledButtons;
    }
    else if (treeState == MULTIPLE_ROWS_SELECTED)
      buttons = multipleRowsEnabledButtons;
  }
  // Enable/disable the toolbar buttons based on the toolbar array
  for (i in toolbarButtons) {
    toolbarButtons[i].disabled = !buttons[i];
  }
  // Additional checks to disable one or more toolbar buttons
  var isLastBrowserWindow = (getNumberOfWindowsByType("navigator:browser") == 1);
  var disabled = ((treeState == NO_ROWS_SELECTED && !isLastBrowserWindow) ||
                  treeState & MULTIPLE_ROWS_SELECTED);
  toolbarButtons[TAB_BUTTON].disabled = disabled;
  toolbarButtons[SESSION_BUTTON].disabled = (treeState == NO_ROWS_SELECTED ||
                                            (treeState == SINGLE_ROW_SELECTED && !(treeState & ROW_IS_CONTAINER)) ||
                                            (treeState == MULTIPLE_ROWS_SELECTED && !(treeState & ROW_IS_CONTAINER)));
  var columnID = "tm-location";
  columnID = ('columns' in tree) ? tree.columns[columnID] : columnID;
  var location = (tree.currentIndex != -1) ? tree.view.getCellText(tree.currentIndex, columnID) : "";
  disabled = (treeState != SINGLE_ROW_SELECTED || (treeState & ROW_IS_CONTAINER));
  toolbarButtons[INTERVAL_BUTTON].disabled = (disabled && location == "");
  toolbarButtons[PROPERTIES_BUTTON].disabled = disabled;
  toolbarButtons[CLOSE_BUTTON].disabled = (treeState == NO_ROWS_SELECTED || (treeState & ROW_IS_CONTAINER && isLastBrowserWindow));
}
////////////////////////////////////////////////////////////////////////////////
function initContextMenu(aEvent)
{
  // Is the context menu item array initialized?
  if (typeof contextMenuItems[0] == 'string') {
    // Nope, not yet, initialize array now
    for (i in contextMenuItems) {
      contextMenuItems[i] = document.getElementById(contextMenuItems[i]);
    }
  }
  for (i in contextMenuItems) {
    contextMenuItems[i].setAttribute('disabled', 'true');
  }
  var treeState = tabManager.getCurrentTreeState();

  if (treeState == NO_ROWS_AVAILABLE)
    return;

  var treeRows = tree.treeBoxObject.view.rowCount;

  if (treeState == MULTIPLE_ROWS_SELECTED) {
    contextMenuItems[CM_SELECT_ALL].setAttribute('disabled', 'true');
    contextMenuItems[CM_SELECT_NONE].removeAttribute('disabled');
  }
  else { // NO_ROWS_SELECTED || SINGLE_ROW_SELECTED
    if (treeRows > 1)
      contextMenuItems[CM_SELECT_ALL].removeAttribute('disabled');
    contextMenuItems[CM_SELECT_NONE].setAttribute('disabled', 'true');
  }
  var columnID = "tm-location";
  columnID = ('columns' in tree) ? tree.columns[columnID] : columnID;
  var location = (tree.currentIndex != -1) ? tree.view.getCellText(tree.currentIndex, columnID) : "";
  var disabled = (treeState != SINGLE_ROW_SELECTED || (treeState & ROW_IS_CONTAINER) || location == "");

  if (!disabled)
    contextMenuItems[CM_INTERVAL].removeAttribute('disabled');
}
//////////////////////////////////////////////////////////////////////////////
function initReloadPopup(aPopup, aEvent)
{
  if (toolbarButtons[3].disabled) {
    aEvent.preventDefault();
    return;
  }
  var tabBrowser = getTabbrowser(true);
  tabBrowser.initReloadPopup(aPopup, null, tabManager.getReloadInterval());
}
////////////////////////////////////////////////////////////////////////////////
function openTMPropertyWindow()
{
  var treeState = tabManager.getCurrentTreeState();

  if (treeState == NO_ROWS_SELECTED || treeState == MULTIPLE_ROWS_SELECTED || currentRow == -1)
    return;
  gResource = tabManager.getResourceForRow(currentRow); // tab    -> urn:session-x:window-x:tabs:tab-x
                                                            // window -> urn:session-x:window-x:tabs:root
  var winFeatures = "chrome, dependent, modal, dialog" + (gUseStackedWindows ? "" : ", resizable=yes, centerscreen");
  window.openDialog("chrome://multiviews/content/tab-manager/properties/tabManagerProperties.xul",
                    "tm-properties", winFeatures, window);
}
////////////////////////////////////////////////////////////////////////////////
function openPrefOverlay()
{
  var winFeatures = "chrome, dependent, modal, dialog" + (gUseStackedWindows ? "" : ", resizable=yes, centerscreen");
  window.openDialog("chrome://multiviews/content/tab-manager/preferences/tabManagerPrefWindow.xul",
                    "tm-preferences", winFeatures, window, "");
}
////////////////////////////////////////////////////////////////////////////////
// This is a dummy function, called from the pref manager (for compatibility only)
function switchTree(gPanel)
{
  return;
}
////////////////////////////////////////////////////////////////////////////////
function switchPanel(aPanel)
{
  var categoryButton = null;
  var panelFile = null;

  prefWindow.gSecondLevelPrefs = false;

  if (aPanel == "tabManagerPrefWindow")
    panelFile = "tabManagerMainPanel.xul";
  prefWindow.refreshPanel(panelFile, categoryButton);
}
////////////////////////////////////////////////////////////////////////////////
function initWindowOpenType()
{
  const prefString = "multizilla.managers.use-stacked-windows";

  try {
    gUseStackedWindows = pref.getBoolPref(prefString);
  } catch(ex) {
    gUseStackedWindows = /Win/.test(navigator.platform);
    pref.setBoolPref(prefString, gUseStackedWindows);
  }
}
////////////////////////////////////////////////////////////////////////////////
function getConfirmation(aWindow, aTitle, aText, aCheckboxText)
{
  gConfirmCheckbox = false;
  var checkbox = {};
  var checkboxText = aCheckboxText ? gStringBundle.GetStringFromName(aCheckboxText) : null;
  var result = gPromptService.confirmEx(aWindow, gStringBundle.GetStringFromName(aTitle), gStringBundle.GetStringFromName(aText),
                                        (gPromptService.BUTTON_POS_0 * gPromptService.BUTTON_TITLE_YES) +
                                        (gPromptService.BUTTON_POS_1 * gPromptService.BUTTON_TITLE_NO),
                                        null, null, null, checkboxText, checkbox);

  gConfirmCheckbox = checkbox.value;
  return (result == 0); // Yes returns 0 / No, Esc and X returns 1
}
////////////////////////////////////////////////////////////////////////////////
function initTreeTooltip(aEvent)
{
  var node = document.tooltipNode;
  var nodeName = node.localName;
  var tooltipNode = aEvent.target;

  if (nodeName == 'tree' || nodeName.match('menu') || nodeName == 'treecol') {
    aEvent.preventDefault();
    return;
  }
  tooltipNode.firstChild.hidden = (nodeName != 'treecols');
  tooltipNode.childNodes[1].hidden = (nodeName != 'splitter');
  tooltipNode.lastChild.hidden = (nodeName != 'treechildren');
}
////////////////////////////////////////////////////////////////////////////////
var tabManager =
{
  windowManagerDS : null,
  tree : null,
  currentRow : null,
  reloadInterval : 0,

  initTabManager: function()
  {
    windowManagerDS = Components.classes['@mozilla.org/rdf/datasource;1?name=window-mediator'].getService(nsIWindowDataSource);
    dataSource = Components.classes['@mozilla.org/rdf/datasource;1?name=in-memory-datasource'].createInstance(Components.interfaces.nsIRDFDataSource);
    tree = document.getElementById('TabManagerTree');
    tree.database.AddDataSource(dataSource);

    gTabbrowser = getTabbrowser(true);
    gStringBundle = Components.classes["@mozilla.org/intl/stringbundle;1"]
                              .getService(Components.interfaces.nsIStringBundleService)
                              .createBundle("chrome://multiviews/locale/multizilla.properties");
    currentRow = -1;
    reloadInterval = 0;
    var prefService = Components.classes["@mozilla.org/preferences-service;1"].getService(Components.interfaces.nsIPrefService);
    pref = prefService.getBranch(null);
    prefPanelNames = gStringBundle.GetStringFromName('prefPanelNames').split(',');

    var prefValue = gSelectedTabStyle;
    const prefString = "multizilla.tab-manager.selectedtab-styling";
    /*** 
      * The pref setting might be missing, but only after updating MultiZilla! This happens 
      * when you forgot to set 'multizilla.prefs.checked' to true, which you should do after 
      * each MultiZilla update btw, but we should still be able to startup without JS errors!
      */
    try {
      prefValue = pref.getComplexValue(prefString, Components.interfaces.nsISupportsString).data;
    } 
    catch(ex) {
      var str = Components.classes["@mozilla.org/supports-string;1"].createInstance(Components.interfaces.nsISupportsString);
      str.data = prefValue;
      pref.setComplexValue(prefString, Components.interfaces.nsISupportsString, str);
    }
    gSelectedTabStyle = prefValue;

    tabManager.buildTree();
    sizeToContent();
    initWindowOpenType();
    updateToolbarButtons();

    gObserverService.notifyObservers(window.document.documentElement, "MZManagers", "start");
    gObserverService.addObserver(gTabUpdateObserver, "tabUpdate", false);
    gWindowManager.addListener(gWindowManagerListener);
  },

  buildTree: function()
  {
    var windowResources = getWindowResourceIDList(true);

    if (windowResources) {
      var sessionID = 0;

      for (i in windowResources) {
        var windowID = windowResources[i];
        var browserWindow = windowManagerDS.getWindowForResource(windowID);
        var tabbrowser = browserWindow.getBrowser();
        var resource = "urn:" + windowID + ":tabs:root";
        gWindowResources.push(resource);
        storeGroupData(dataSource, resource, tabbrowser, true, false, gSelectedTabStyle);
      }
    }
  },

  refreshTree: function()
  {
    for (i in gWindowResources) {
      var resource = gWindowResources[i];
      var windowID = resource.replace(/urn:/, '').replace(/:tabs:root/, '');
      var browserWindow = windowManagerDS.getWindowForResource(windowID);

      if (browserWindow) {
        var tabbrowser = browserWindow.getBrowser();
        storeGroupData(dataSource, resource, tabbrowser, true, false, gSelectedTabStyle);
      }
    }
  },

  click: function(aEvent)
  {
    currentRow = this.getRowNumber(aEvent);

    if (currentRow == -1)
      return;
    reloadInterval = this.getReloadInterval(aEvent);
  },

  getRowNumber: function(aEvent)
  {
    var X = aEvent.pageX;
    var Y = aEvent.pageY;
    return tree.treeBoxObject.getRowAt(X, Y);
  },

  getResourceForRow: function(aRowNumber)
  {
    return (aRowNumber > -1) ? tree.builderView.getResourceAtIndex(aRowNumber).Value : null;
  },

  getSelectedRows: function()
  {
    var rows = -1;
    var selectedRows = new Array();
    // Get and return all selection ranges
    var rangeCount = tree.builderView.selection.getRangeCount();
    // Iterate over all ranges to collect the selected rows
    for (var range = 0; range < rangeCount; range++) {
      var startSelection = {};
      var endSelection = {};
      tree.builderView.selection.getRangeAt(range, startSelection, endSelection);
      // Collect all rows and store them in our array
      for (var row = startSelection.value; row <= endSelection.value; row++) {
        // Check for invalid row numbers
        if (startSelection.value != -1 || endSelection.value != -1)
          selectedRows[++rows] = row;
      }
    }
    if (selectedRows)
      return selectedRows;
    return null;
  },

  getCurrentTreeState: function()
  {
    if (tree.treeBoxObject.view.rowCount < 1)
      return NO_ROWS_AVAILABLE;

    var selectedRows = this.getSelectedRows();

    if (selectedRows.length > 1) {
      if (this.isMixedSelection())
        return (MULTIPLE_ROWS_SELECTED + MIXED_ROWS_SELECTED);
	  else if (tree.treeBoxObject.view.isContainer(Number(selectedRows[0])))
        return (MULTIPLE_ROWS_SELECTED + ROW_IS_CONTAINER);
      return MULTIPLE_ROWS_SELECTED;
    }
    else if (selectedRows.length == 1) {
      if (tree.treeBoxObject.view.isContainer(Number(selectedRows[0])))
        return (SINGLE_ROW_SELECTED + ROW_IS_CONTAINER);
      return SINGLE_ROW_SELECTED;
    }
    return NO_ROWS_SELECTED;
  },

  getRowSource: function(row, inout)
  {
    var resource = this.getResourceForRow(row);
    var windowID = resource.replace(/urn:/, '');

	if (tree.treeBoxObject.view.isContainer(row))
      windowID = windowID.replace(/:tabs:\w*/, '');
    else
      windowID = windowID.replace(/:tabs:tab-\w*/, '');

    var browserWindow = windowManagerDS.getWindowForResource(windowID);
    var tabbrowser = browserWindow.getBrowser();

    var columnID = "tm-ordinal";
    columnID = ('columns' in tree) ? tree.columns[columnID] : columnID;
    var ordinal = tree.view.getCellText(row, columnID);
    var tab = tabbrowser.mTabList[ordinal];
    var browser = tabbrowser.getBrowserForTab(tab);

    inout.window = browserWindow;
    inout.windowID = windowID;
    inout.resource = resource;
    inout.tabbrowser = tabbrowser;
    inout.tab = tab;
    inout.browser = browser;
  },

  getReloadInterval: function(aEvent)
  {
    var interval = 0;
    var columnID = "tm-interval";
    columnID = ('columns' in tree) ? tree.columns[columnID] : columnID;

    if (aEvent) {
      var row = this.getRowNumber(aEvent);
      interval = (row > -1) ? tree.view.getCellText(row, columnID) : 0;
    }
    else if (!aEvent && tree.currentIndex != -1)
      interval = tree.view.getCellText(tree.currentIndex, columnID);

    if (interval)
      return interval;
    return 1;
  },

  setReloadInterval: function(aEvent, aValue)
  {
    var resource = this.getResourceForRow(currentRow);
    var intervalProp = RDF.GetResource("http://multizilla.mozdev.org/rdf#interval");
    var currentValue = dataSource.GetTarget(RDF.GetResource(resource), intervalProp, true);

    if (aValue == 0) // Remove interval property from tree
      dataSource.Unassert(RDF.GetResource(resource), intervalProp, currentValue);
    else
      addOrChangeValue(dataSource, RDF.GetResource(resource), intervalProp, currentValue, RDF.GetLiteral(aValue));

    mzFlushDataSource(dataSource);
    this.updateTab("interval", aValue);
  },

  isMixedSelection: function()
  {
    var tabs = 0;
    var windows = 0;
    var mixed = false;

    // Collect selected tree rows
    var selectedItems = this.getSelectedRows();
    // Do we have any selected items?
    if (selectedItems && selectedItems.length > 0) {
      // Iterate over the seleted items
      for (var row = 0; row < selectedItems.length; row++) {
        if (tree.treeBoxObject.view.isContainer(selectedItems[row]))
          windows++;
        else
          tabs++;
        // Do we have a mixed selection?
        if (tabs > 0 && windows > 0) {
          mixed = true;
          break;
        }
      }
    }
    return mixed;
  },

  removeWindowItemsFromSelection: function()
  {
    // Collect selected tree rows
    var selectedRows = this.getSelectedRows();
    // Do we have any selected items?
    if (selectedRows) {
      // Iterate over the selected items to clear the 'window' (container) items
      for (row in selectedRows) {
        if (tree.treeBoxObject.view.isContainer(row))
          tree.view.selection.clearRange(row, row);
      }
    }
  },

  selectRows: function(aCommand)
  {
    switch (aCommand) {
      case "all": // Called from the context menu, or Ctrl+A ?
           tree.view.selection.selectAll();
           this.removeWindowItemsFromSelection();
           break;
      case "none": // Called from the context menu, or Ctrl+Z ?
           tree.view.selection.clearSelection();
           // dump("\ntree.currentIndex: " + tree.currentIndex);
           tree.view.selection.select(tree.currentIndex);
           tree.treeBoxObject.ensureRowIsVisible(tree.currentIndex);
           break;
    }
    tree.focus();
  },

  activateSelectedTab: function(row)
  {
    var obj = new Object();
    this.getRowSource(row, obj);

    if (obj.resource.indexOf("root") != -1) {
      document.commandDispatcher.advanceFocusIntoSubtree(obj.tabbrowser.selectedTab);
    }
    else {
      obj.tabbrowser.selectedTab = obj.tab;
      document.commandDispatcher.advanceFocusIntoSubtree(obj.tab);
    }
    this.refreshAfterCommand();
  },

  add: function(openType)
  {
    if (openType == 'window') {
      window.open();
    }
    else {
      var treeState = this.getCurrentTreeState();

      if (treeState == NO_ROWS_SELECTED) {
        if (openType == 'tab') {
          var tabbrowser = getTabbrowser(true);
          tabbrowser.addTab();
        }
      }
      else {
        var obj;

        if (openType == 'tab') {
          obj = new Object();
          this.getRowSource(currentRow, obj);
          obj.tabbrowser.addTab();
        }
        else if (openType == 'session') {
          if (mzRDFStore == null)
            initializeRDFStorageFile();

          if (!this.isMixedSelection()) {
            var selectedRows = this.getSelectedRows();

            if (selectedRows) {
              var windowID = 'window-0';
              var sessionID = 0;
              var currentWindow = 0;
              var skippedWindows = 0;

              for (var i = 0; i < selectedRows.length; i++)
              {
                if (tree.treeBoxObject.view.isContainer(Number(selectedRows[i])))
                {
                  obj = new Object();
                  this.getRowSource(selectedRows[i], obj);

                  if (areAllTabsBlank(obj.tabbrowser)) {
                    skippedWindows++;
                    continue; // Yes, skip this window
                  }
                  /***
                    * obj.windowID might be set to 'window-10' but that still doesn't mean it is the 
                    * tenth open browser window and that is why we need an indexed value of windowID
                    */
                  if (currentWindow == 0) {
                    sessionID = getNextSessionID(sessionID, windowID);
                  }
                  else {
                    windowID = 'window-' + currentWindow;
                  }
                  var selectedTabStyle;
                  const prefString = "multizilla.session-manager.selectedtab-styling";

                  try {
                    selectedTabStyle = pref.getComplexValue(prefString, Components.interfaces.nsISupportsString).data;
                  } catch(ex) {
                    selectedTabStyle = Components.classes["@mozilla.org/supports-string;1"].createInstance(Components.interfaces.nsISupportsString);
                    selectedTabStyle.data = "bold";
                    pref.setComplexValue(prefString, Components.interfaces.nsISupportsString, selectedTabStyle);
                  }
                  var resource = "urn:" + sessionID + ":" + windowID + ":tabs:root";
                  storeGroupData(mzRDFStore, resource, obj.tabbrowser, true, false, selectedTabStyle);
                  currentWindow++;
                }
              }
              var totalWindows = currentWindow;
              var promptTitle, promptText;
              var notificationArray = new Array();

              if (skippedWindows) {
                promptTitle = gStringBundle.GetStringFromName('tmAddSessionNotificationTitle');
                notificationArray = [ skippedWindows, totalWindows ];
                promptText = gStringBundle.formatStringFromName('tmSkippedWindowsText', notificationArray, 2);
                gPromptService.alert(window, promptTitle, promptText);
              }
              /***
                * Get rid of the tree selection(s). This should prevent people from 
                * adding the same session over and over again.
                */
              tree.view.selection.clearSelection();
              tree.currentIndex = -1;

              if (totalWindows) {
                promptTitle = gStringBundle.GetStringFromName('tmAddSessionNotificationTitle');
                sessionID = sessionID.replace(/session-/, '');
                notificationArray = [ totalWindows, sessionID ];
                promptText = gStringBundle.formatStringFromName('tmSavedWindowsText', notificationArray, 2);
                gPromptService.alert(window, promptTitle, promptText);
              }
            }
          }
        }
        obj = null;
      }
    }
    this.refreshAfterCommand();
  },

  closeTabs: function(allTabs)
  {
    if (allTabs) // needed for context menu items, might be moved out
      tabManager.selectRows('all');
    this.closeSelectedItems();
  },

  closeWindows: function()
  {
    this.closeSelectedItems();
  },

  closeSelectedItems: function()
  {
    var treeState = this.getCurrentTreeState();

    if (treeState == NO_ROWS_SELECTED)
      return;
    else {
      var confirmRequired = true;
      var closeTabsConfirmation, closeWindowsConfirmation;
      const confirmPrefID = "multizilla.tab-manager.confirm.close-";
      // read the pref, and add it, with the default value, when it is missing
      try {
        closeTabsConfirmation = pref.getBoolPref(confirmPrefID + "tabs");
      } catch(ex) {
        closeTabsConfirmation = true;
        pref.setBoolPref(confirmPrefID + "tabs", closeTabsConfirmation);
      }
      try {
        closeWindowsConfirmation = pref.getBoolPref(confirmPrefID + "windows");
      } catch(ex) {
        closeWindowsConfirmation = true;
        pref.setBoolPref(confirmPrefID + "windows", closeWindowsConfirmation);
      }
      if (confirmRequired) {
        var header = 'tabs-windows';

        if (treeState & SINGLE_ROW_SELECTED) {
          header = (treeState & ROW_IS_CONTAINER) ? 'window' : 'tab';
        }
        else { // MULTIPLE_ROWS_SELECTED
          if (!(treeState & MIXED_ROWS_SELECTED))
            header = (treeState & ROW_IS_CONTAINER) ? 'windows' : 'tabs';
        }
        var confirmationTitle = header + '.confirmDialogTitle';
        var confirmationText = header + '.confirmDialogText';
        // Ask for the users confirmation before removing the tab(s) and/or window(s)
        if ((closeTabsConfirmation && header.match(/tab/)) ||
            (closeWindowsConfirmation && header.match(/window/))) {

          if (!getConfirmation(window, confirmationTitle, confirmationText, null)) {
            tree.focus();
            return;
          }
        }
      }
      this.close();
    }
  },

  close: function()
  {
    mzDebug("close()");
    var selectedRows = this.getSelectedRows();

    if (selectedRows) {
      // Close the tab(s) or Window(s) one after one, from buttom up!
      for (var i = (selectedRows.length-1); i > -1; i--) {
        var obj = new Object();
        this.getRowSource(selectedRows[i], obj);
        // is this a window 'urn:window-x:tabs:root'
        if (obj.resource.indexOf("root") != -1) {
          gWindowResources.pop(obj.resource);
          removeAllResourcesFor(dataSource, obj.resource); // remove window from tree
          obj.window.close();
        }
        else { // this must be a tab 'urn:window-x:tabs:tab-x'
          // is this the last tab?
          if (obj.tabbrowser.mTabContainer.childNodes.length > 1)
            removeResources(dataSource, obj.resource); // no, remove tab from tree
          obj.tabbrowser.removeTab(obj.tab);
        }
      }
      obj = null;
      mzFlushDataSource(dataSource);
      tree.builder.rebuild();
      this.refreshAfterCommand();
    }
    mzDebug("exit closeTabs()");
  },

  refreshAfterCommand: function()
  {
    /* var focusedWindow = document.commandDispatcher.focusedWindow;
    if (focusedWindow != window)
      dump("\nWrong Window Focused"); */

    updateToolbarButtons();
    // window.focus(); // focus TM Window, to make the ESC key work
    tree.focus(); // focus TM Tree, to make key up/down work
    document.commandDispatcher.focusedWindow = window;
    document.commandDispatcher.focusedElement = tree;
  },

  updateTab: function(updateType, data)
  {
    dump("\nupdateTab");
    var treeState = this.getCurrentTreeState();

    if (treeState != SINGLE_ROW_SELECTED || tree.currentIndex == -1)
      return;

    var obj = new Object();
    this.getRowSource(tree.currentIndex, obj);
    var tab = obj.tab;
    var browser = obj.browser;
    var tabbrowser = obj.tabbrowser;
    var columnID;

    switch(updateType) {
      case "label":
        tab.setAttribute("label", data);
        columnID = "Name";
        columnID = ('columns' in tree) ? tree.columns[columnID] : columnID;
        tree.builderView.setCellText(currentRow, columnID, data);
        break;
      case "customlabel":
        if (data)
          tab.setAttribute("customLabel", "true");
        else
          tabbrowser.removeCustomLabel();
        // tree.builderView.setCellValue(currentRow, '', data);
        // tree.builderView.setCellText(currentRow, '', data);
        break;
      case "location":
        browser.loadURI(data);
        columnID = "tm-location";
        columnID = ('columns' in tree) ? tree.columns[columnID] : columnID;
        tree.builderView.setCellText(currentRow, columnID, data);
        break;
      case "interval":
        tabbrowser.setReloadInterval(tab, data);
        break;
      case "allowPlugins":
      case "allowJavascript":
      case "allowMetaRedirects":
      case "allowSubframes":
      case "allowImages":
        browser.docShell[updateType] = data;
        tabbrowser.refreshDocShellPropertyButtons();
        break;
    }
  },

  onSelect: function()
  {
    currentRow = tree.currentIndex;
    updateToolbarButtons();
    var treeState = this.getCurrentTreeState();

    if (currentRow != -1 && (treeState & SINGLE_ROW_SELECTED))
      this.activateSelectedTab(currentRow);
  },

  onHelp: function(aPageAnchor)
  {
    var pageAnchor = "";
    var helpPages = gStringBundle.GetStringFromName("helpPages-TabManager");

    if (aPageAnchor) {
      if (typeof aPageAnchor == 'string')
        pageAnchor = "#" + aPageAnchor;
      else // event
        pageAnchor = "#" + aPageAnchor.target.id.replace(/^tabManager-/, '');
    }
    // We could eventually check the on-line state!
    var tabbrowser = getTabbrowser(true);
    tabbrowser.selectedTab = tabbrowser.addOrGetTab(helpPages + pageAnchor);
    window.close();
  },

  onCancel: function()
  {
    var retValue = this.confirmClose(gConfirmForID);
    return retValue;
  },

  confirmClose: function(aConfirm)
  {
    var shouldClose = true;
    const confirmPrefID = "multizilla.tab-manager.confirm.quit";
    var dialogCheckboxText, confirmPrefValue;
    /***
      * 1  = Close Button
      * 2  = ESC key pressed
      * 4  = Ctrl-W
      * 8  = Click on [X] on Window
      */
    try {
      /*** 
        * The pref setting might be missing, but only after updating MultiZilla! This happens 
        * when you forgot to set 'multizilla.prefs.checked' to true, which you should do after 
        * each MultiZilla update btw, but confirmClose() should still be fully functional!
        */
      confirmPrefValue = pref.getIntPref(confirmPrefID);
    } catch(ex) {
      confirmPrefValue = 15;
      pref.setIntPref(confirmPrefID, confirmPrefValue);
    }
    // Is the pref set for confirmation?
    if (confirmPrefValue && (confirmPrefValue & aConfirm) == aConfirm) {
      // Yes, ask the users confirmation
      const checkboxText = 'managerSkipConfirmFor';

      switch(aConfirm) {
        case (aConfirm & 1): dialogCheckboxText = checkboxText + 'CloseButton';
          break;
        case (aConfirm & 2): dialogCheckboxText = checkboxText + 'EscapeKey';
          break;
        case (aConfirm & 4): dialogCheckboxText = checkboxText + 'CtrlW';
          break;
        case (aConfirm & 8): dialogCheckboxText = checkboxText + 'WindowX';
          break;
      }
      shouldClose = getConfirmation(window, 'tmConfirmCloseTitle', 'tmConfirmCloseText', dialogCheckboxText);
      // Is the checkbox checked and do we have a pref value to work with?
      if (gConfirmCheckbox && (confirmPrefValue >= aConfirm)) {
        // Yes, update pref
        pref.setIntPref(confirmPrefID, (confirmPrefValue -= aConfirm));
      }
    }
    if (shouldClose) {
      gObserverService.notifyObservers(window.document.documentElement, "MZManagers", "stop");
      gObserverService.removeObserver(gTabUpdateObserver, "tabUpdate");
      gWindowManager.removeListener(gWindowManagerListener);
    }
    return shouldClose;
  }
};

addEventListener("load", tabManager.initTabManager, false);

/*
////////////////////////////////////////////////////////////////////////////////
function updateCurrentIndex()
{
  tree.currentIndex = gTabbrowser.selectedTab.ordinal;
  tree.view.selection.toggleSelect(tree.currentIndex);
}

var tabManagerDNDObserver =
{
  tree : null,

  initTabManagerDNDObserver: function()
  {
    tree = document.getElementById("tabManagerTree");
    initializeRDFStorageFile();
    tree.database.AddDataSource(mzRDFStore);
    tree.builder.rebuild();
    updateCurrentIndex();
    tabbrowser = getLastActiveBrowserWindow().getBrowser();
    stringBundle = document.getElementById("tm-properties");
    tabListInit();
  },

  onDragStart: function(aEvent, aXferData, aDragSession)
  {
    // start the row drag operation only when there are at least 2 visible rows
    if (tree.treeBoxObject.view.rowCount > 1) {
      aEvent.target.setAttribute("dragover", "true");
      var columnID = "sm-interval";
      columnID = ('columns' in tree) ? tree.columns[columnID] : columnID;
      var ordinal = tree.view.getCellText(tree.currentIndex, columnID);
      // dump("onDragStart ordinal: "+ordinal+"\n");
      aXferData.data = new TransferData();
      aXferData.data.addDataForFlavour("text/unicode", ordinal);
    }
  },

  onDragOver: function(aEvent, aXferData, aDragSession)
  {
    var X = aEvent.pageX;
    var Y = aEvent.pageY;
    var row = tree.treeBoxObject.getRowAt(X, Y);

    aDragSession.canDrop = true;
    if (row > -1) {
      dump("onDragOver - row: "+row+"\n");
      var columnID = "sm-interval";
      columnID = ('columns' in tree) ? tree.columns[columnID] : columnID;
      var ordinal = tree.view.getCellText(row, columnID);
      dump("ordinal: "+ordinal+"\n");
      ordinal++;
      var element = document.getElementById("tm-tab-"+ordinal);
      dump(element.nodeName+"\n");
      // nsDragAndDrop.mDragSession.sourceNode = element;
    }
  },

  onDrop: function(aEvent, aXferData, aDragSession)
  {
    dump("onDrop\n"+aEvent.originalTarget.nodeName);
  },

  onDragExit: function(aEvent, aXferData, aDragSession)
  {
    tree.childNodes[1].removeAttribute("dragover");
  },

  getSupportedFlavours: function ()
  {
    // we might want to expand this
    var flavourSet = new FlavourSet();
    flavourSet.appendFlavour("text/unicode");
    return flavourSet;
  },

  canDrop: function (aEvent, aDragSession)
  {
    aDragSession.canDrop = true;
    // allow drops from the Tab Manager only
    var doc = aDragSession.sourceNode.ownerDocument;
    var elem = doc.getElementById("tabManagerWindow");
    return (elem && (elem.getAttribute("windowtype") == "multizilla:tabmanager"));
  }
};
addEventListener("load", tabManagerDNDObserver.initTabManagerDNDObserver, false); */

